<?php
if (!defined('MEDIAWIKI')) die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * Class to handle queries of the WikiDB.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2012, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 663 $
 */

WikiDB_RegisterRevision('$Rev: 663 $');

class WikiDB_Query {
	var $pDB;
	var $pErrorMsg;

	var $pTables = array();
	var $pSortFields = array();
	var $pCriteriaFields = array();

/////////////// PUBLIC FUNCTIONS ///////////////

// WikiDB_Query() - Constructor
// Takes a table string, and optional criteria and sort strings, which are used to
// set up the query object ready for running.
// When the object has been initialised, call the Run() function to perform the 
// actual query and return the full set of results.
	function WikiDB_Query($TableString, $CriteriaString = "", $SortString = "") {
		$this->pDB =& wfGetDB(DB_SLAVE);
		$this->pErrorMsg = "";

		if (!$this->pSetTablesFromString($TableString))
			return;
		if (!$this->pSetCriteriaFromString($CriteriaString))
			return;
		if (!$this->pSetSortFromString($SortString))
			return;
	}

// Run()
// Runs the defined query and returns an array containing the results, one item per
// row, indexed numerically.
// If an error occurs, a string is returned containing the error message.
	function Run() {
		if ($this->pErrorMsg != "")
			return $this->pErrorMsg;

		$this->pGenerateFieldLists();
		return $this->pPerformQuery();
	}

/////////////// SETTING UP INPUTS ///////////////

// pSetTablesFromString()
// Takes a table string and sets up the array of tables used in this query.
// Currently a table string may only contain a single table.  In future we may
// allow multiple tables, or joins.
// Table aliases are resolved before we add the data to our array.
// The tables array is indexed by the full table name, and each element contains the
// necessary data pertaining to the given table (the table object, the name of the
// alias that will be used in the query, and an array of referenced fields, initially
// empty).
// TODO: Implement multi-table queries.
	function pSetTablesFromString($TableString) {
	// Create a table object for this table, resolving any redirects to get the
	// actual table that the data ends up in.
	// The query that we build will ensure that all aliases are also checked.
		$Table = new WikiDB_Table($TableString);
		$Table = $Table->GetDataDestination();

	// Get canonical name of table
		$TableName = $Table->GetFullName();
		
	// If this table is already set up, then we don't need to define it a second
	// time.
		if (isset($this->pTables[$TableName]))
			return true;

	// Create an alias for this table, which will be used to reference it within the
	// query (as the query may contain multiple WikiDB tables, each built from the
	// same set of DB tables).
	// The alias will be used as the name for this table's instance of the rowdata 
	// table, and as a prefix to the field name when naming each instance of the
	// fielddata table that relates to this table.
	// We simply use a numeric counter for this, with the first table being called
	// "table1".
		$Alias = "table" . (count($this->pTables) + 1);

	// Create the entry for this table, initially with an empty field list.
	// The field list will be added to when we parse the other query parameters,
	// and will ultimately contain a list of all fields in this table that are
	// referenced by the query (and therefore which need to be included in the
	// JOIN expression).
		$this->pTables[$TableName]['table'] = $Table;
		$this->pTables[$TableName]['alias'] = $Alias;
		$this->pTables[$TableName]['fields'] = array();

	// Return true to indicate success.
		return true;
	}

// pSetCriteriaFromString()
// Parses the criteria string, and sets up an array of criteria data indicating
// the involved fields and the operator(s) that apply.
// This function is currently very crude, because the allowed syntax is very
// crude - this all needs expanding.
// TODO: Allow for multiple tables.
// TODO: Expand, expand, expand!
//
// The current criteria string may contain the following:
// 	* Multiple criteria are allowed, separated by commas.  These will be ANDed
//	  together when the query is run (currently no way to OR criteria).
//	* Each criteria is of the form "FIELD OPERATOR VALUE", e.g. "Artist=Madonna".
//	* Whitespace is allowed in the criteria string, and is automatically removed.
//
//	* FIELD is of the form "TABLE.FIELDNAME", where "TABLE." is optional if there is
//	  only one table being used in the query.  If more than one table is being used,
//	  then it is required for all fields (not applicable currently).
//  * TABLE aliases are resolved properly.
//	* The resulting TABLE (after alias resolution) must match one of the tables 
//    defined for the query.
//	* FIELDNAME does not have to be defined in the table, as undefined fields may
//    still contain data.
//	* Field aliases are resolved properly if FIELDNAME exists in the table 
//	  definition.
//
//	* OPERATOR is any of the valid operators listed in $ValidOperators.
//
//	* VALUE is the value for comparison, unquoted.  Currently fields cannot be
// 	  compared to each other.
//
	function pSetCriteriaFromString($CriteriaString) {
		$ValidOperators = array("<>", ">=", "<=", "!=", ">", "<", "=");

	// Setup regular expression for matching individual criteria.
		$Regex = "";
		foreach ($ValidOperators as $Operator) {
			if (!IsBlank($Regex))
				$Regex .= "|";
			$Regex .= preg_quote($Operator);
		}
		$Regex = '/^(.*)((?-U)' . $Regex . ')(.*)$/iU';

		$WhereFields = array();

//		print("-> " . $CriteriaString . "<br>");

	// Criteria are comma-separated, so explode on commas to get the individual criteria.
		foreach (explode(",", $CriteriaString) as $Criteria) {
//			print("* " . $Criteria . "<br>");
			if(preg_match($Regex, $Criteria, $Matches)) {
//				print_r($Matches);
				$Field = trim($Matches[1]);
				$Value = trim($Matches[3]);
				$Operator = $Matches[2];

			// Modify the != operator to use the <> operator, as the former is not
			// recognised by MySQL.
				if ($Operator == "!=")
					$Operator = "<>";

			// Parse the field to get the proper table/field names.
			// All aliases are resolved.
				$FieldData = $this->pParseFieldName($Field);

			// If an error occurred, then return without doing anything more.
				if ($FieldData == false)
					return false;

				$Value = $FieldData['table']->FormatFieldForSorting(
														$FieldData['fieldname'],
														$Value
													);

			// The Value parameter is required, so if it is not set ignore this
			// criterion.
			// TODO: Need to be able to compare against "" (and/or NULL).
				if ($Value != "") {
					$WhereFields[] = array(
											'table' => $FieldData['tablename'],
											'field' => $FieldData['fieldname'],
											'op' => $Operator,
											'value' => $Value,
										);
				}
			}
		}

		$this->pCriteriaFields = $WhereFields;

	// Return true to indicate success.
		return true;
	}

// pSetSortFromString()
// Parses the sort string, and sets up an array of sort data indicating the fields
// to sort on and the direction of sort.
// TODO: Allow for multiple tables.
//
// The sort string may contain the following:
//	* Multiple sort criteria are allowed, separated by commas.  The resulting data
//	  will be sorted by each of these fields in order, in the same way as a standard
//	  ORDER BY clause.
//	* Each criterion is of the form "FIELD DIRECTION", where DIRECTION is optional.
//	* Whitespace is allowed and ignored.
//
//	* FIELD follows the same rules as for the WHERE criteria, above.
//	* DIRECTION is either "ASC" or "DESC" (case-insensitive).  If omitted, "ASC"
//    is used.
//
	function pSetSortFromString($SortString) {
		$SortFields = array();

		foreach (explode(",", $SortString) as $SortField) {
			if(preg_match('/^(.*)(?:\s(ASC|DESC))?\s*$/iU', $SortField, $Matches)) {
				$Field = trim($Matches[1]);
				if ($Field == "")
					continue;

			// Normalise the $Direction property, setting it if omitted.
			// We don't need to check the value is valid, as the regex has already
			// done that for us.
				if (!isset($Matches[2]))
					$Direction = "ASC";
				else
					$Direction = strtoupper($Matches[2]);

			// Parse the field to get the proper table/field names.
			// All aliases are resolved.
				$FieldData = $this->pParseFieldName($Field);

			// If an error occurred, then return without doing anything more.
				if ($FieldData == false)
					return false;

				$SortFields[] = array(
									'table' => $FieldData['tablename'],
									'field' => $FieldData['fieldname'],
									'dir' => $Direction,
								);
			}
		}

		$this->pSortFields = $SortFields;

	// Return true to indicate success.
		return true;
	}

// pParseFieldName()
// Parses a string to get the field name.  If the field string contains a period
// then the portion prior to the period is the name of a table, otherwise the table
// defined for the query is assumed (and if more than one is defined, an error is
// raised).
// If the string contains a table component that wasn't in the query's table
// string then an error is raised.  Note that table aliases are resolved before this
// check.
// If valid, an array containing the 'tablename' and 'fieldname' is returned.  The
// array also contains the WikiDB_Table object for the table, to save calling 
// functions having to recreate it, if they need access to it for any reason.
// Otherwise pError is set and false is returned.
// Table aliases are resolved, and so are field aliases, if the field exists in 
// the table definition.
	function pParseFieldName($FieldString) {
	// Get position of period in field name
		$DotPos = strpos($FieldString, ".");

	// If there is no period in the field string, then use the table defined in the
	// query's table string.
		if ($DotPos === false) {
		// If there is more than one table defined for this query, then the field
		// string is ambiguous, so raise an error.
			if (count($this->pTables) > 1) {
				$this->pError = "Ambiguous fields in WHERE clause.";
				return false;
			}

		// Otherwise, use the table defined in the query's table string.
		// This has already had aliases resolved.
			$FullTableName = key($this->pTables);
			$Table = $this->pTables[$FullTableName]['table'];
		}

	// If the field string contains a period then it contains a table name
	// component, which needs separating and checking.
		else {
		// Extract table name from Field
			$TableName = trim(substr($FieldString, 0, $DotPos - 1 ));
		// Set $FieldString to hold just the field component.
			$FieldString = trim(substr($FieldString, $DotPos + 1));

		// Resolve table aliases, and resolve table name to it's canonical version
		// (i.e. including namespace).
			$Table = new WikiDB_Table($TableName);
			$Table = $Table->GetDataDestination();
			$FullTableName = $Table->GetFullName();

		// If the table is not defined for use in the query, raise an error.
			if (!isset($this->pTables[$FullTableName])) {
				$this->pError = "Undefined table '" . $TableName
							  . "' in WHERE clause.";
				return false;
			}
		}

	// Resolve field aliases if the field is defined.
		if ($Table->FieldIsDefined($FieldString))
			$FieldString = $Table->GetAlias($FieldString);

		return array(
					'tablename' => $FullTableName, 
					'fieldname' => $FieldString,
					'table'		=> $Table,
				);
	}

/////////////// GENERATING INTERMEDIATE DATA ///////////////

// pGenerateFieldLists()
// Fills in the 'fields' array for each table.  This will become an array of fields 
// in that table which are referenced in the query for filtering or sorting purposes
// (i.e. the fields that we need to include in the output).  
// The array is indexed by the actual field name, with the value being the name
// we will use for the table when creating the join statement required for that 
// field.
// All table and field aliases have been resolved prior to calling this function.
	function pGenerateFieldLists() {
		$FieldNames = array();

	// Create an array of all fields used in the query.
	// We simply merge the sort and criteria arrays.  Although the structure of the
	// elements in each array is different, we are only interested in the 'table'
	// and 'field' elements, which are always present in both arrays.
		$arrAllFields = array_merge($this->pSortFields, $this->pCriteriaFields);
		
	// Reset the two field lists, as we will rebuild them during the subsequent loop.
	// We do it this way, because we want to add an 'alias' field, yet we have no
	// way to link the record in the merged array back to the source array.  It is
	// therefore simplest just to wipe the source arrays and re-add the individual
	// items as we process them.
		$this->pSortFields = array();
		$this->pCriteriaFields = array();
	
	// Set up the aliases.
		foreach ($arrAllFields as $Field) {
		// Create some short-cuts to make code more readable.
			$TableName = $Field['table'];
			$FieldName = $Field['field'];

		// Create the table alias that will be used for this field.  Each field
		// requires a separate join to the WikiDB fields table, and this is the name
		// that will be used.
			$Alias = $this->pTables[$TableName]['alias'] . "_" . $FieldName;
			$Alias = $this->pDB->tableName($Alias);

		// Store the $FieldName => $Alias mapping into the list of fields used by
		// this table, so that they are created as part of the JOIN.
			$this->pTables[$TableName]['fields'][$FieldName] = $Alias;
		
		// Also add the alias to the original field definition, so it can be used
		// in the WHERE/ORDER BY clauses.
		// We know that if the item has a 'dir' element, then this is a sort field,
		// otherwise it is a criteria field.
			$Field['tablealias'] = $Alias;
			if (isset($Field['dir']))
				$this->pSortFields[] = $Field;
			else
				$this->pCriteriaFields[] = $Field;
		}
	}

/////////////// PERFORMING THE QUERY ///////////////

// pPerformQuery()
// Generates and runs the query, returning the full set of matched data rows
// in the standard format.
// TODO: Allow multi-table queries.
	function pPerformQuery() {
	// Block multiple tables for now.
		if (count($this->pTables) != 1) {
			$this->pErrorMsg = "Currently only queries involving a single table "
							 . "are supported.";
			return;
		}

		reset($this->pTables);
		$TableData = $this->pTables[key($this->pTables)];
		$Query = $this->pGetQuerySQL($TableData, $this->pCriteriaFields,
									 $this->pSortFields);
		$Result = $this->pDB->query($Query);

		$RowData = array();
		$RowID = 0;
		while ($Row = $this->pDB->fetchObject($Result)) {
		// Extract the data for this row.
			$RowData[$RowID]['Data'] = unserialize($Row->parsed_data);
		
		// Add the meta-data.
			$RowData[$RowID]['MetaData']['_Row'] = $RowID + 1;
			$RowData[$RowID]['MetaData']['_SourceArticle'] = 
								WikiDB_Table::GetFullName($Row->page_title, 
														  $Row->page_namespace);

			$RowID++;
		}
		$this->pDB->freeResult($Result);

//		print_r($RowData);
		return $RowData;
	}

// pGetQuerySQL()
// Generate the SQL statement used by pPerformQuery().
	function pGetQuerySQL($TableData, $WhereFields, $SortFields) {
	// Create some short-cuts, for tidier code.
		$Table = $TableData['table'];
		$FieldList = $TableData['fields'];
		$DB =& $this->pDB;

	// Get the correct table names for the real tables (properly escaped).
		$RowTable = $DB->tableName(pWIKIDB_tblRows);
		$FieldTable = $DB->tableName(pWIKIDB_tblFields);

	// Escape row table's alias.  We do this as we intend to allow for multiple
	// WikiDB tables to be used in one query (even though at the moment it is only
	// possible to have one).
		$RowTableAlias = $DB->tableName($TableData['alias']);

	// Create field list.
	// We only need the namespace title and data for the row, which comes directly
	// from the row table.  The data for individual fields, which is in the fields
	// table, are only used in the WHERE and ORDER BY clauses and so don't need
	// to be part of the field list.
		$SQL = "SELECT " 
		     . $RowTableAlias . ".page_namespace, "
			 . $RowTableAlias . ".page_title, "
			 . $RowTableAlias . ".parsed_data";

	// Create JOIN expression.
	// We join the table row table to multiple instances of the field table, one
	// for each field that is used in either the WHERE or ORDER BY clause.
	// Note that when stepping through the field lists, the $FieldName is not yet
	// escaped, and so always needs quoting, whereas the $FieldTableName has already
	// been escaped properly.  The quoting for $FieldName differs as in some places
	// it is used as a column name, where in others it is a string to be matched.
		$SQL .= " FROM " . $RowTable . " AS " . $RowTableAlias;
		foreach ($FieldList as $FieldName => $FieldTableAlias) {
			$FieldString = "";
			$Aliases = $Table->GetAliasRev($FieldName);
	 
			foreach ($Aliases as $Alias) {
				if (!IsBlank($FieldString))
					$FieldString .= " OR ";
				$FieldString .= $FieldTableAlias . ".field_name = "
							  . $DB->addQuotes($Alias);
			}

			$SQL .= " LEFT JOIN " . $FieldTable . " AS " . $FieldTableAlias
				  . " ON " . $FieldTableAlias . ".row_id = "
				  . $RowTableAlias . ".row_id" . " AND (" . $FieldString . ")";
		}

	// Create WHERE clause to restrict to the required table.
	// We need to compare the table name against all valid aliases.
		$SQL .= " WHERE (";
		foreach ($Table->GetTableAliases() as $Index => $TableAlias) {
			if ($Index > 0)
				$SQL .= " OR ";

			$SQL .= "(" . $RowTableAlias . ".table_namespace = "
				  . $DB->addQuotes($TableAlias->GetNamespace())
				  . " AND " . $RowTableAlias . ".table_title = "
				  . $DB->addQuotes($TableAlias->GetDBKey())
				  . ")";
		}
		$SQL .= ")";

	// Add criteria specified for query, if there are any.
		if (count($WhereFields) > 0) {
			$SQL .= " AND (";
			foreach ($WhereFields as $i => $Criteria) {
				if ($i > 0)
					$SQL .= " AND ";
			
				$SQL .= $Criteria['tablealias'] . ".field_value"
					  . " " . $Criteria['op'] . " "
					  . $DB->addQuotes($Criteria['value']);
			}
			$SQL .= ")";
		}

	// Add ORDER BY clause if there is one
		if (count($SortFields) > 0) {
			$SQL .= " ORDER BY ";
			foreach ($SortFields as $i => $Sort) {
				if ($i > 0)
					$SQL .= ", ";

				$SQL .= $Sort['tablealias'] . ".field_value";
				if ($Sort['dir'] == "DESC")
					$SQL .= " DESC";
			}
		}

		$SQL .= ";";

		return $SQL;
	}

} // END: class WikiDB_Query

