<?php
/**
 * RebuildWikiDB.php
 * Rebuilds the WikiDB cache, by reparsing all articles in the wiki.
 * Because of this it can be very slow, particularly on large databases!
 * Arguments:
 * 		--start		If specified, start from this article ID, rather than article 1.
 *  	--wipefirst	If specified, the whole cache is wiped first.  This is currently
 * 				 	the only way to delete cached data for deleted articles.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2012, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 661 $
 */

$Usage = <<<EOF
Rebuilds the whole WikiDB cache by re-parsing all pages in the database.
This can be very slow, but in normal use is safe to interrupt
You can specify a start article using --start=N.
To ensure deleted articles are removed from the cache, use --wipefirst.
The --skip-refresh argument can be used to bypass the code which updates 
any stale records created due to data definitions changing.

EOF;
//////////////////////////////////////

	$optionsWithArgs = array('start');

// Include header file required for all WikiDB maintenance scripts.
	require_once(dirname( __FILE__ ) . "/header.inc");
	
	$Start = null;
	if (isset($options['start'])) {
	// If --start is not a valid positive integer then die with an error.
		if (!preg_match('/^\d+$/', $options['start']))
			die("Invalid value for --start argument: " . $options['start'] . "\n");
		
		$Start = intval($options['start']);
	}

// Give a warning, as this can be a very slow script on large wikis.
	print("About to update the WikiDB data cache, which involves reparsing ALL pages!\n");
	print("Warning: This can be a very slow process!\n");
	AllowUserToCancel();
	
// If requested, wipe the table first.
	if (isset($options['wipefirst'])) {
		print("'wipefirst' specified - clearing cache completely...");
		ClearAllWikiDBCaches();
		print("done.\n");
	}

// Prepare DB
	wfWaitForSlaves(false);
	$dbr =& wfGetDB(DB_SLAVE);

// Setup parser
	# Don't generate TeX PNGs (lack of a sensible current directory causes errors anyway)
	# Built-in Maths support was removed in MW 1.18, so check for existence of the
	# constant before attempting to use it.
	if (defined("MW_MATH_SOURCE")) {
		$wgUser->setOption('math', MW_MATH_SOURCE);
	}

	# Don't generate thumbnail images
	$wgUseImageResize = false;
	$wgUseTidy = false;

// Get page IDs
	$SQL = "SELECT page_id FROM " . $dbr->tableName('page');
// Skip pages in the MediaWiki namespace.
	$SQL .= " WHERE page_namespace <> " . NS_MEDIAWIKI;
// If a start ID was specified, begin from that record.
	if (!is_null($Start))
		$SQL .= " AND page_id >= " . $Start;
// Order them by page_id so that we can interrupt and resume intelligently.
	$SQL .= " ORDER BY page_id";

	$res = $dbr->query($SQL, __FILE__);

// Parse all pages
	$i = 0;
	$LastReport = "";
	$RowCount = $dbr->affectedRows();

	if (!is_null($Start))
		print("Skipping pages with a page_id lower than " . $Start . ".\n");
	print("Updating " . $RowCount . " pages: ");

	while ($row = $dbr->fetchObject($res)) {
		if ((++$i % pWIKIDB_ReportingInterval) == 0) {
			print_c_fixed($LastReport, $i);
			$LastReport = $i;
		}

		ReparseArticle($row->page_id);
	}
	
	print_c_fixed($LastReport, "Complete!\n");

// Update stale data rows.  If '--skip-refresh' option specified then we only
// report how many are stale.
	if (isset($options['skip-refresh'])) {
		print("Fixing of stale records skipped: ");
		$options['report'] = true;
	}
	include(dirname(__FILE__) . "/RefreshStaleData.php");
