<?php
/**
 * RefreshStaleData.php
 * Field data can go stale if the the table definition they belong to gets updated.
 * This is because we do not automatically update all data when the table changes,
 * because for tables with a lot of data this would grind the wiki to a halt.  
 * Instead, when a table definition changes, we mark all affected rows as stale and 
 * then update them in batches on each page request.  
 * This script allows you to manually update stale records from the command-line.
 * Arguments:
 *  	--report	This simply reports the number of stale records and exits.
 *  	--limit=N	Allows you to perform a batch update of N revisions.  If omitted
 *  				then all stale data is updated.  This switch is useful if you
 *  				run the script via a regular cron-job and want to avoid 
 *  				excessive load.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2012, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 661 $
 */

$Usage = <<<EOF
Refreshes any stale field data that has gone out of date to changes to its table definition.  By default the script refreshes all stale data.  Use --limit=N to restrict this.  If you specify the --report switch, then the script will simply report the number of stale records and exit.

EOF;
//////////////////////////////////////

	$optionsWithArgs = array('limit');

// Include header file required for all WikiDB maintenance scripts.
	require_once(dirname( __FILE__ ) . "/header.inc");

// Setup the limit based on the --limit argument (if present).
	$Limit = 0;
	if (isset($options['limit'])) {
	// If --limit is not a valid positive, non-zero integer then die with an error.
		if (!preg_match('/^\d+$/', $options['limit']) || $options['limit'] < 1)
			die("Invalid value for --limit argument: " . $options['limit'] . "\n");
		
		$Limit = intval($options['limit']);
	}

// Retrieve and report the current number of stale links.
	$StaleRows = $DB->selectField(pWIKIDB_tblRows, "COUNT(*)", 
								   array("is_stale", true), __FILE__);
	print($StaleRows . " stale data row");
	if ($StaleRows != 1)
		print("s");
	print(" in DB.");

// If there are no stale links, or the --report flag was set, that's all we need to
// do!
	if ($StaleRows == 0 || isset($options['report'])) {
		print("\n");
		return;
	}

// If limit is greater than the number of links to refresh then we don't need
// a limit.  (I am assuming no race conditions here!)
	if ($Limit >= $StaleRows)
		$Limit = 0;

// Report on what we're going to do.
	print("  ");
	if ($Limit === 0)
		print("All");
	else
		print($Limit);
	print(" stale row");
	if ($Limit != 1)
		print("s");
	print(" will be refreshed.\n");

	WikiDB::RefreshStaleFieldData($Limit, true);

// Report on remaining stale rows.  Note that this is re-counted from the DB, and it
// is possible that new rows have become stale since the script started.
	$StaleRows = $DB->selectField(pWIKIDB_tblRows, "COUNT(*)", 
								   array("is_stale", true), __FILE__);

	print($StaleRows . " stale row");
	if ($StaleRows != 1)
		print("s");
	print(" remaining.\n");